#include <OS.h>
#include <Drivers.h>
#include <KernelExport.h>
#include <SupportDefs.h>

#include "mm2000.h"
#include "ISA.h"

#define TIMEOUT 10000  // max polling before fail

struct dev_inf
{
	int32  address;
};

_EXPORT
int32	api_version = B_CUR_DRIVER_API_VERSION;


static int32 open_count = 0;

/*
 *   driver-related structures
 */
 
const char *mm_name[] = 
{
 	"misc/mm2000",
 	NULL
};
          

/*
 *   init_driver - set debug mode and find isa base
 */ 

char  *ISA_name = B_ISA_MODULE_NAME;

isa_module_info *isa;

_EXPORT
status_t init_driver(void)
{
    set_dprintf_enabled( FALSE );

    dprintf("mm2000 driver: %s %s\n", __DATE__, __TIME__);

	return get_module(ISA_name, (module_info **)&isa);
}

_EXPORT
void uninit_driver(void)
{
	put_module(ISA_name);
}

int32 lookup_device_name (const char *name)
{
	int i;
	
	for (i = 0; mm_name[i]; i++)
		if (!strcmp (mm_name[i], name))
			return i;
	return -1;
}


/*
 *   mm_open - called on open().
 */
status_t mm_open( const char *name, uint32 flags, void **cookie )
{
     struct dev_inf* dv;


     *cookie = (void*)calloc( 1, sizeof(struct dev_inf) );

     dv = (struct dev_inf*)*cookie;

     dv->address = 0;  // null address to start with

     atomic_add( &open_count, 1);
     return B_NO_ERROR;
}


/*
 *   mm_close
 */

status_t mm_close( void *dev )
{

     struct dev_inf* dv;

     dv = (struct dev_inf*)dev;


     if( dev ) free( dev );

     atomic_add( &open_count, -1 );
     return B_NO_ERROR;
}

/*
 *   mm_free - don't know what it's for  maybe we free dev here instead?
 *
 */
status_t mm_free( void *dev )
{
     return B_NO_ERROR;
}



/*
 *   mm_read - read some stuff from our device.  //pos arg may be used later to write counter
 *
 */

status_t mm_read( void *dev, off_t pos, void *buf, size_t *len )
{
     struct dev_inf* dv;
	 uint32 i=0;
     uint32 j;
     uchar *rsp;
     int32 cnt_down;

     dv = (struct dev_inf*)dev;

     rsp = buf;

	 j = *len;

     if( ! dv->address ) return(-1);  // return an error if address isn't set yet

     do {
          cnt_down=TIMEOUT;
          while(--cnt_down != 0)
          {
               if( isa->read_io_8( dv->address+1 ) == 0 ) break;
          }
          if( cnt_down == 0 )  break;

          isa->write_io_8( dv->address+1, 128 );

          cnt_down=TIMEOUT;
          while( --cnt_down != 0 )
          {
               if( isa->read_io_8( dv->address+1 ) >= 128 ) break;
          }

          if( cnt_down == 0 ) break;

          rsp[i++] = isa->read_io_8( dv->address );

          isa->write_io_8( dv->address+1, 0 );

     }
     while( rsp[i-1] != 10 && cnt_down != 0 && i < j );
        
     if( cnt_down != 0 )
     {
          cnt_down=TIMEOUT;
          while (--cnt_down != 0)
          {
               if( isa->read_io_8( dv->address+1 ) == 128) break;
          }
     }
	 
	 if( i > 1 ) i -= 2;  // be sure we dont get seg violation
     else i=0;

     rsp[i] = '\0';  // null terminate the reply

     if( cnt_down == 0 ) return -1;         
     else return 1;
}


 /*
 *   mm_write - write some stuff to our device. //pos arg may be used later to write counter
 */

status_t mm_write( void *dev, off_t pos, const void *buf, size_t *len )
{
     struct dev_inf* dv;
     uint32 j;
     uint32 i;
     const char *cmd;
     int32 cnt_down;
     cpu_status status;
     dv = (struct dev_inf*)dev;

     i = 0;
	 j = *len;
     cmd = buf;

    if( ! dv->address ) return(-1);

     status = disable_interrupts();
 
     while( i < j )
     {
          restore_interrupts( status );
          cnt_down = TIMEOUT;
          while( --cnt_down > 0 )
          {
               if( isa->read_io_8( dv->address+1 ) == 128 ) break;
          }

          if( cnt_down == 0 ) break;


          isa->write_io_8( dv->address, cmd[i++] );
          isa->write_io_8( dv->address+1, 128 );

          cnt_down = TIMEOUT;
          while( --cnt_down > 0 )
          {
               if( isa->read_io_8( dv->address+1 ) == 0 ) break;
          }


          if( cnt_down == 0 ) break;

          status = disable_interrupts();
          isa->write_io_8( dv->address+1, 0 );
     }

     if( cnt_down != 0 )
     {
          cnt_down = TIMEOUT;
          while( --cnt_down > 0 )
          {
               if( isa->read_io_8( dv->address+1 ) == 128 ) break;
          }
      }

      restore_interrupts( status );

      if( cnt_down == 0 ) return 0;
      else return i;
}


 /*
  *   isa_control
  */

status_t mm_control( void *dev, uint32 msg, void *buf, size_t len )
{
     struct dev_inf* dv;
     int32  *ap;

     dv = (struct dev_inf*)dev;
     ap = buf;

     switch( msg )
     {

          case B_GET_DEVICE_SIZE:
               break;
          case MM2000_LOAD_ADDRESS:
               dv->address = *ap;
               break;
           default:
               return B_ERROR;
     }

     return B_NO_ERROR;
}

_EXPORT
device_hooks mm2000_device = 
{
	&mm_open,           /* -> open entry point */
	&mm_close,          /* -> close entry point */
	&mm_free,
	&mm_control,   /* -> control entry point */
	&mm_read,      /* -> read entry point */
	&mm_write,      /* -> write entry point */
          NULL,
          NULL
};


_EXPORT
const char **publish_devices()
{
	return mm_name;
}

_EXPORT
device_hooks *find_device(const char *name)
{
	return &mm2000_device;
}
